
/**
	@file libtfshm.c
	@brief 簡易共有メモリライブラリ
	@author Atsushi WATANEBE <http://www.roboken.esys.tsukuba.ac.jp/~atusi_w/>
 */
 // TODO: System V message queue から POSIX message queue に変更

#define _XOPEN_SOURCE
#define _BSD_SOURCE
#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>

#include <sys/file.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/time.h>
#include <sys/types.h>

#include <errno.h>
#include <limits.h>
#include <unistd.h>

#include <tfshm.h>

#include "readwrite-nbyte.h"
#if ( defined( S_IRUSR ) && defined( S_IWUSR ) )
# define _S_DEFAULT		S_IRUSR | S_IWUSR
#else
# define _S_DEFAULT		0
#endif
#if ( defined( S_IRGRP ) && defined( S_IWGRP ) )
# define S_DEFAULT		_S_DEFAULT | S_IRGRP | S_IWGRP
#else
# define S_DEFAULT		_S_DEFAULT
#endif
unsigned short name2num( char *name )
{
	char *p;
	unsigned short num;
	num = 0;
	for ( p = name; *p; p++ )
	{
		num = ( num << 3 ) | ( ( num >> ( 16 - 3 ) ) & 0xF );
		num ^= *p;
	}
	return num;
}

/**
	@brief Open shared memory server
	@param *name Name of server
	@param *path Path to the shm
	@return IO number of port
 */
tfshm *tfshm_openex( char *name, char *path )
{
	tfshm *port;
	char portname[LIBTFSHM_NAME_LENGTH];

	port = malloc( sizeof ( tfshm ) );

	if( port )
	{
		port->param = 0;
		port->data = 0;

		if( path == NULL )
			path = LIBTFSHM_DEFAULT_PATH;

		snprintf( portname, LIBTFSHM_NAME_LENGTH, "%s/%s.tfshm.data", path, name );
		port->data = open( portname, O_RDONLY );
		if( port->data <= 0 )
		{
			tfshm_close( port );
			return NULL;
		}
		// printf( "%s = %d\n", name, name2num(name) );
		port->event = msgget( name2num( name ), 0666 );
		strcpy( port->name, name );
		strcpy( port->path, path );

		snprintf( portname, LIBTFSHM_NAME_LENGTH, "%s/%s.tfshm.param", path, name );
		port->param = open( portname, O_RDONLY );
		if( port->param <= 0 )
		{
			tfshm_close( port );
			return NULL;
		}

		port->did = 0;
	}
	else
	{
		return NULL;
	}

	return port;
}

/**
	@brief Open shared memory server
	@param *name Name of server
	@return IO number of port
 */
tfshm *tfshm_open( char *name )
{
	return tfshm_openex( name, NULL );
}

/**
	@brief Create shared memory server
	@param *name Name of server
	@param *path Path to the shm
	@return IO number of port
 */
tfshm *tfshm_createex( char *name, char *path )
{
	tfshm *port;
	char portname[LIBTFSHM_NAME_LENGTH];

	port = malloc( sizeof ( tfshm ) );

	if( port )
	{
		long msg;

		if( path == NULL )
			path = LIBTFSHM_DEFAULT_PATH;

		snprintf( portname, LIBTFSHM_NAME_LENGTH, "%s/%s.tfshm.data", path, name );
		port->data = open( portname, O_WRONLY | O_CREAT, S_DEFAULT );
		if( port->data <= 0 )
		{
			tfshm_close( port );
			return NULL;
		}

		// printf( "%s = %d\n", name, name2num(name) );
		port->event = msgget( name2num( name ), 0666 | IPC_CREAT );

		strcpy( port->name, name );
		strcpy( port->path, path );

		port->did = 0;
		while( 1 )
		{
			if( msgrcv( port->event, &msg, 0, -( LONG_MAX - 1 ), IPC_NOWAIT ) == -1 )
			{
				break;
			}
		}
		msg = 1;
		if( msgsnd( port->event, &msg, 0, 0 ) == -1 )
		{
			close( port->data );
			printf( "W msgsnd failed.\n" );
			return NULL;
		}

		snprintf( portname, LIBTFSHM_NAME_LENGTH, "%s/%s.tfshm.param", path, name );
		port->param = open( portname, O_WRONLY | O_CREAT, S_DEFAULT );
		if( port->param <= 0 )
		{
			tfshm_close( port );
			return NULL;
		}
	}

	return port;
}

/**
	@brief Create shared memory server
	@param *name Name of server
	@return IO number of port
 */
tfshm *tfshm_create( char *name )
{
	return tfshm_createex( name, NULL );
}

/**
	@brief Receive data list
	@param port IO number of port
	@param **data Pointer to pointer to data buffer
	@param size Size of buffer
	@param autoalloc malloc automaticaly
	@param ctl Control structure
	@param block Blocking or not
	@return Read size
 */
size_t tfshm_readex( tfshm * port, void **data, size_t size, int autoalloc, int block, tfshm_control * ctl )
{
	int ret;
	struct flock fl;
	long msg;

	fl.l_whence = SEEK_SET;						/* SEEK_SET, SEEK_CUR, SEEK_END */
	fl.l_start = 0;								/* Offset from l_whence */
	fl.l_len = 0;								/* length, 0 = to EOF */
	fl.l_pid = getpid(  );						/* our PID */

	// Check event
	if( block )
	{
		if( msgrcv( port->event, &msg, 0, -( LONG_MAX - ( port->did + 1 ) ), 0 ) == -1 )
		{

			close( port->data );
			printf( "R msgrcv failed.\n" );
			return -1;
		}
		if( msgsnd( port->event, &msg, 0, 0 ) == -1 )
		{

			close( port->data );
			printf( "R msgsnd failed.\n" );
			return -1;
		}
		if( msg == 1 )
		{
			port->did = 0;
			if( msgrcv( port->event, &msg, 0, 1, MSG_EXCEPT ) == -1 )
			{
				close( port->data );
				printf( "R msgrcv failed.\n" );
				return -1;
			}
			if( msgsnd( port->event, &msg, 0, 0 ) == -1 )
			{

				close( port->data );
				printf( "R msgsnd failed.\n" );
				return -1;
			}
		}
		else
		{
			port->did = ( LONG_MAX - msg );
		}
		// printf( "%d\n",(int)port->did );
	}

	// Lock data
	fl.l_type = F_RDLCK;
	if( fcntl( port->data, F_SETLKW, &fl ) == -1 )
	{

		close( port->data );
		printf( "R DATA lock failed.\n" );
		return -1;
	}

	// Seek
	lseek( port->data, SEEK_SET, 0 );

	// Read size
	ret = readn( port->data, ctl, sizeof ( tfshm_control ) );
	if( ret <= 0 )
	{
		size = 0;
		goto unlock_read;
	}
	// Read
	if( autoalloc )
	{
		if( ctl->size > size )
		{
			if( *data )
				free( *data );
			*data = malloc( ctl->size );
			size = ctl->size;
		}
	}
	if( size > ctl->size )
		size = ctl->size;
	ret = readn( port->data, *data, size );
	if( ret <= 0 )
	{
		size = 0;
		goto unlock_read;
	}

  unlock_read:
	// Unlock
	fl.l_type = F_UNLCK;
	if( fcntl( port->data, F_SETLKW, &fl ) == -1 )
	{

		close( port->data );
		printf( "R DATA unlock failed.\n" );
		return -1;
	}
	return size;
}

/**
	@brief Send data list
	@param port IO number of port
	@param *data Pointer to data buffer
	@param size Size of buffer
	@param ctl Control structure
	@return Written size
 */
size_t tfshm_writeex( tfshm * port, void *data, tfshm_control * ctl )
{
	int ret;
	struct flock fl;
	long msg;

	fl.l_whence = SEEK_SET;
	fl.l_start = 0;
	fl.l_len = 0;
	fl.l_pid = ctl->pid;

	// Lock data
	fl.l_type = F_WRLCK;
	if( fcntl( port->data, F_SETLKW, &fl ) == -1 )
	{

		close( port->data );
		printf( "W DATA lock failed.\n" );
		return -1;
	}

	// Seek
	lseek( port->data, SEEK_SET, 0 );

	// Write size
	ret = writen( port->data, ctl, sizeof ( tfshm_control ) );
	if( ret <= 0 )
	{
		ctl->size = 0;
		goto unlock_write;
	}
	// Write
	ret = writen( port->data, data, ctl->size );
	if( ret <= 0 )
	{
		ctl->size = 0;
		goto unlock_write;
	}

	// Clear event
	while( 1 )
	{
		if( msgrcv( port->event, &msg, 0, -( LONG_MAX - 1 ), IPC_NOWAIT ) == -1 )
		{
			break;
		}
	}
	// Notify event
	msg = LONG_MAX - ( port->did + 1 );
	port->did++;
	if( msgsnd( port->event, &msg, 0, 0 ) == -1 )
	{

		close( port->data );
		printf( "W msgsnd failed.\n" );
		return -1;
	}

  unlock_write:
	// Unlock
	fl.l_type = F_UNLCK;
	if( fcntl( port->data, F_SETLKW, &fl ) == -1 )
	{

		close( port->data );
		printf( "W DATA unlock failed.\n" );
		return -1;
	}
	return ctl->size;
}

/**
	@brief Receive data list
	@param port IO number of port
	@param *data Pointer to data buffer
	@param size Size of buffer
	@return Read size
 */
size_t tfshm_read( tfshm * port, void *data, size_t size )
{
	tfshm_control ctl;

	return tfshm_readex( port, &data, size, 0, 1, &ctl );
}

/**
	@brief Receive data list
	@param port IO number of port
	@param *data Pointer to data buffer
	@param size Size of buffer
	@return Read size
 */
size_t tfshm_read_nonblock( tfshm * port, void *data, size_t size )
{
	tfshm_control ctl;

	return tfshm_readex( port, &data, size, 0, 0, &ctl );
}

/**
	@brief Get parameter
	@param port IO number of port
	@param *data Pointer to parameter buffer
	@param size Size of buffer
	@return Read size
 */
size_t tfshm_getparam( tfshm * port, void *data, size_t size )
{
	size_t ret;
	// struct flock fl;

	/* 
	 * fl.l_whence = SEEK_SET; fl.l_start = 0; fl.l_len = 0; fl.l_pid = getpid();
	 * 
	 * // Lock data fl.l_type = F_RDLCK; if( fcntl( port->param, F_SETLKW, &fl ) == -1 ){
	 * 
	 * close( port->param ); printf("R PARAM lock failed.\n"); return -1; } */
	// Seek
	lseek( port->param, SEEK_SET, 0 );

	// Read
	ret = readn( port->param, data, size );
	if( ret <= 0 )
	{
		ret = 0;
		// goto unlock_read_param;
	}
	else
	{
		ret = size;
	}

	/* 
	 * unlock_read_param: //Unlock fl.l_type = F_UNLCK; if( fcntl( port->param, F_SETLKW, &fl ) == -1 ){
	 * 
	 * close( port->param ); printf("R PARAM unlock failed.\n"); return -1; } */
	return ret;
}

/**
	@brief Send data list
	@param port IO number of port
	@param *data Pointer to point buffer
	@param size Size of buffer
	@return Written size
 */
size_t tfshm_write( tfshm * port, void *data, size_t size )
{
	tfshm_control ctl;

	ctl.size = size;
	ctl.reserved[0] = 0;
	ctl.pid = getpid(  );
	gettimeofday( &ctl.timestamp, NULL );

	return tfshm_writeex( port, data, &ctl );
}

/**
	@brief Put patrameter
	@param port IO number of port
	@param *data Pointer to parameter buffer
	@param size Size of buffer
	@return Written size
 */
size_t tfshm_putparam( tfshm * port, void *data, size_t size )
{
	int ret;
	struct flock fl;

	fl.l_whence = SEEK_SET;
	fl.l_start = 0;
	fl.l_len = 0;
	fl.l_pid = getpid(  );

	// Lock data
	fl.l_type = F_WRLCK;
	if( fcntl( port->param, F_SETLKW, &fl ) == -1 )
	{

		close( port->param );
		printf( "W PARAM lock failed.\n" );
		return -1;
	}

	// Seek
	lseek( port->param, SEEK_SET, 0 );

	// Write
	ret = writen( port->param, data, size );
	if( ret <= 0 )
	{
		ret = 0;
		goto unlock_write_param;
	}

  unlock_write_param:
	// Unlock
	fl.l_type = F_UNLCK;
	if( fcntl( port->param, F_SETLKW, &fl ) == -1 )
	{

		close( port->data );
		printf( "W DATA unlock failed.\n" );
		return -1;
	}
	return ret;
}

/**
	@brief Free
	@param port IO number of port
 */
void tfshm_close( tfshm * port )
{
	if( port->param > 0 )
	{
		close( port->param );
		port->param = 0;
	}
	if( port->data > 0 )
	{
		close( port->data );
		port->data = 0;
	}
	free( port );

}
