/**
	@file tfshm-amanuensis.c
	@brief 簡易共有メモリライブラリのログ記録ユーティリティ
	@author Atsushi WATANEBE <http://www.roboken.esys.tsukuba.ac.jp/‾atusi_w/>
 */
 
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <pthread.h>
#include <poll.h>
#include <sys/file.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <signal.h>
#include <time.h>
#include <limits.h>
#include <termios.h>
#include <tfshm.h>
#include "tfshm-amanuensis.h"
#include "readwrite-nbyte.h"
	
void nano_usleep( time_t usec )
{
	struct timespec req;
	req.tv_nsec = usec * 1000;
	req.tv_sec = req.tv_nsec / 1000000000;
	req.tv_nsec = req.tv_nsec % 1000000000;
	nanosleep( &req, NULL );
}

void nano_gettimeofday( struct timeval *res, void *na )
{
	struct timespec res2;
	clock_gettime( CLOCK_REALTIME, &res2 );
	res->tv_sec = res2.tv_sec;
	res->tv_usec = res2.tv_nsec / 1000;
}

struct timeval g_simulated_time;
struct timeval g_start_time;
pthread_mutex_t g_simulated_time_mutex;
int g_shutoff;
int g_num;
float g_bai;
int g_shmwriten[1024];
int g_shmsync[1024];
int g_shmdelay[1024];
int g_shmend[1024];
int g_shmwait[1024];
char *g_shmname[1024];

/**
	@brief 引数の説明表示
	@param argc 引数数
	@param argv 引数
 */
int arg_help( int argc, char *argv[] )
{
	fprintf( stderr, "\nUSAGE: %s [OPTION]...\n\n", argv[0] );
	fprintf( stderr, "  -r, --record <dir>    Record shared data..\n" );
	fprintf( stderr, "  -p, --play <dir>      Play recorded data..\n" );
	fprintf( stderr, "  -m, --shmpath <path>  Specify the path of shared memory..\n" );
	fprintf( stderr, "  -h, --help            This help\n\n" );
	
	return 1;
}

/**
	@brief Ctrl+C trap
	@param aN not used
 */
void ctrlc( int aN )
{
	g_shutoff = 1;
	signal( SIGINT, NULL );
}

/**
	@brief 引数の説明表示
	@param argc 引数数
	@param argv 引数
 */
int main( int argc, char *argv[] )
{
	int i;
	char *logfile = NULL;
	char *shmpath = LIBTFSHM_DEFAULT_PATH;
	enum
	{
		MODE_NONE,
		MODE_PLAY,
		MODE_RECORD
	} mode = MODE_NONE;
	for ( i = 1; i < argc; i++ )
	{
		if( !strcmp( argv[i], "--record" ) || !strcmp( argv[i], "-r" ) )
		{
			i++;
			if( i < argc )
			{
				logfile = argv[i];
				mode = MODE_RECORD;
			}
		}
		else if( !strcmp( argv[i], "--play" ) || !strcmp( argv[i], "-p" ) )
		{
			i++;
			if( i < argc )
			{
				logfile = argv[i];
				mode = MODE_PLAY;
			}
		}
		else if( !strcmp( argv[i], "--shmpath" ) || !strcmp( argv[i], "-m" ) )
		{
			i++;
			if( i < argc )
			{
				shmpath = argv[i];
			}
		}
		else if( !strcmp( argv[i], "--help" ) || !strcmp( argv[i], "-h" ) )
		{
			mode = MODE_NONE;
		}
	}
	// Start trapping ctrl+c signal
	g_shutoff = 0;
	g_num = 0;
	signal( SIGINT, ctrlc );
	// nice( -2 );
	switch ( mode )
	{
	case MODE_NONE:
		return arg_help( argc, argv );
		break;
	case MODE_RECORD:
		return record( logfile, shmpath );
		break;
	case MODE_PLAY:
		return play( logfile, shmpath );
		break;
	}
	return 0;
}

/**
	@brief 情報表示
	@param *arg not used
 */
void *shm_terminal( void *arg )
{
	int shmwriten_last[1024][10];
	int name_len[1024];
	char name_space[1024][64];
	int i;
	int j;
	int name_len_max;
	memset( name_space, ' ', 1024 * 64 );
	name_len_max = 0;
	for ( i = 0; i < g_num; i++ )
	{
		name_len[i] = strlen( g_shmname[i] );
		if( name_len_max < name_len[i] )
			name_len_max = name_len[i];
	}
	for ( i = 0; i < g_num; i++ )
	{
		name_space[i][name_len_max - name_len[i]] = 0;
	}
	j = 0;
	while( !g_shutoff )
	{
		printf( "Time:%5d[sec] x%f\033[0K\n", ( int )( g_simulated_time.tv_sec - g_start_time.tv_sec ), g_bai );
		for ( i = 0; i < g_num; i++ )
		{
			char state;
			if( g_shmend[i] )
			{
				state = 'E';
			}
			else if( g_shmwait[i] )
			{
				state = 'D';
			}
			else
			{
				state = ' ';
			}
			printf( "%c %s[%s] %5d %3dHz  (skipped:%d)\033[0K\n",
					state,
					name_space[i], g_shmname[i], g_shmwriten[i],
					( g_shmwriten[i] - shmwriten_last[i][( j + 10 - 5 ) % 10] ), g_shmdelay[i] );
			shmwriten_last[i][j % 10] = g_shmwriten[i];
		}
		printf( "\033[%dA", g_num + 1 );
		fflush( stdout );
		nano_usleep( 200000 );
		// タイミングをずらして順にHDD書き込み
		g_shmsync[j % g_num] = 1;
		j++;
	}
	printf( "\033[%dB", g_num + 1 );
	nano_usleep( 200000 );
	return NULL;
}

/**
	@brief 記録
	@param dir 保存先ディレクトリ
	@param shmpath shmのpath
 */
int record( char *dir, char *shmpath )
{
	DIR *dp;
	int fd;
	struct timeval starttime;
	pthread_t thread_terminal;
	if( mkdir( dir, S_IRWXU | S_IRWXG ) != 0 )
	{
		fprintf( stderr, "ERROR: Could not create %s.\n", dir );
		return 0;
	}
	chdir( dir );
	dp = opendir( shmpath );
	if( dp == NULL )
	{
		fprintf( stderr, "ERROR: Could not open %s.\n", shmpath );
		return 0;
	}
	fd = open( "starttime", O_WRONLY | O_CREAT | O_EXCL, S_IRUSR | S_IRGRP | S_IWUSR | S_IWGRP );
	if( fd == -1 )
	{
		fprintf( stderr, "ERROR: Could not open starttime.\n" );
		return 0;
	}
	nano_gettimeofday( &starttime, NULL );
	writen( fd, &starttime, sizeof ( starttime ) );
	close( fd );
	rewinddir( dp );
	while( 1 )
	{
		struct dirent *file;
		file = readdir( dp );
		if( file == NULL )
			break;
		if( strstr( file->d_name, ".tfshm.data" ) != NULL && file->d_name[0] != '_' )
		{
			tfshm *shm;
			char name[256];
			char *dot;
			pthread_t thread;
			strcpy( name, file->d_name );
			dot = strchr( name, '.' );
			*dot = 0;
			shm = tfshm_openex( name, shmpath );
			pthread_create( &thread, NULL, record_shm, shm );
			pthread_detach( thread );
		}
		if( strstr( file->d_name, ".tfshm.param" ) != NULL && file->d_name[0] != '_' )
		{
			char cmd[256];
			strcpy( cmd, "cp " );
			strcat( cmd, shmpath );
			strcat( cmd, "/" );
			strcat( cmd, file->d_name );
			strcat( cmd, " ./ > /dev/null 2> /dev/null" );
			system( cmd );
		}
	}
	nano_usleep( 100000 );
	pthread_create( &thread_terminal, NULL, shm_terminal, NULL );
	pthread_join( thread_terminal, NULL );
	closedir( dp );
	return 1;
}

/**
	@brief 1データ記録
	@param *data shm
 */
void *record_shm( void *arg )
{
	tfshm *shm;
	int fd;
	char filename[LIBTFSHM_NAME_LENGTH];
	size_t size = 0;
	void *data = NULL;
	struct timeval before;
	tfshm_control ctl;
	int num;
	long did_old = 0;
	shm = arg;
	snprintf( filename, LIBTFSHM_NAME_LENGTH, "%s.tfshm.log", shm->name );
	num = g_num;
	g_num++;
	g_shmwriten[num] = 0;
	g_shmdelay[num] = 0;
	g_shmend[num] = 0;
	g_shmwait[num] = 0;
	g_shmname[num] = shm->name;
	timerclear( &before );
	timerclear( &ctl.timestamp );
	fd = open( filename, O_WRONLY | O_CREAT | O_EXCL, S_IRUSR | S_IRGRP | S_IWUSR | S_IWGRP );
	if( fd == -1 )
	{
		fprintf( stderr, "ERROR: Could not create %s.\n", filename );
		return NULL;
	}
	while( !g_shutoff )
	{
		before = ctl.timestamp;
		// Blocking & Auto malloc
		if( tfshm_readex( shm, &data, size, 1, 1, &ctl ) > size )
		{
			size = ctl.size;
		}
		if( did_old != 0 && shm->did != 0 )
		{
			if( shm->did - did_old > 1 )
			{
				g_shmdelay[num]++;
			}
		}
		did_old = shm->did;
		writen( fd, &ctl, sizeof ( ctl ) );
		writen( fd, data, ctl.size );
		g_shmwriten[num]++;
		if( g_shmsync[num] )
		{
			fsync( fd );
			g_shmsync[num] = 0;
		}
	}
	fsync( fd );
	free( data );
	tfshm_close( shm );
	close( fd );
	ctl.size = 0;
	ctl.timestamp.tv_sec = 0x7FFFFFFF;
	ctl.timestamp.tv_usec = 0;
	writen( fd, &ctl, sizeof ( ctl ) );
	writen( fd, data, ctl.size );
	return NULL;
}

/**
	@brief 再生
	@param dir 保存先ディレクトリ
	@param shmpath shmのpath
 */
int play( char *dir, char *shmpath )
{
	DIR *dp;
	int fd;
	struct pollfd pfd;
	struct timeval vtime;
	struct timeval zero;
	struct timeval user_start;
	int offset, offset_u, start;
	enum
	{
		MODE_RUN = 0,
		MODE_PAUSE
	} mode;
	int j;
	pthread_t thread_terminal;
	int waiting;
	struct termios old_tios;
	struct termios tios;
	
	dp = opendir( dir );
	if( dp == NULL )
	{
		fprintf( stderr, "ERROR: Could not open %s.\n", dir );
		return 0;
	}
	chdir( dir );
	
	fd = open( "starttime", O_RDONLY );
	if( fd == -1 )
	{
		fprintf( stderr, "ERROR: Could not open starttime.\n" );
		return 0;
	}
	readn( fd, &zero, sizeof ( zero ) );
	close( fd );
	g_start_time = zero;
	start = zero.tv_sec;
	user_start.tv_sec = start;
	user_start.tv_usec = 0;
	
	timerclear( &g_simulated_time );
	tcgetattr( fileno( stdin ), &tios );
	old_tios = tios;
	tios.c_lflag &= ~ECHO;
	tcsetattr( fileno( stdin ), TCSANOW, &tios );
	
	waiting = 0;
	j = 0;
	while( 1 )
	{
		struct dirent *file;
		file = readdir( dp );
		if( file == NULL )
			break;
		if( strstr( file->d_name, ".tfshm.log" ) != NULL )
		{
			tfshm *shm;
			char name[256];
			char *dot;
			pthread_t thread;
			char cmd[256];
			
			strcpy( name, file->d_name );
			dot = strchr( name, '.' );
			*dot = 0;
			shm = tfshm_createex( name, shmpath );
			strcpy( cmd, "cp ./" );
			strcat( cmd, name );
			strcat( cmd, ".tfshm.param " );
			strcat( cmd, shmpath );
			strcat( cmd, " > /dev/null 2> /dev/null" );
			system( cmd );
			pthread_create( &thread, NULL, play_shm, shm );
			pthread_detach( thread );
		}
	}
	closedir( dp );
	
	mode = MODE_RUN;
	pfd.fd = 0;
	pfd.events = POLLIN;
	pthread_mutex_init( &g_simulated_time_mutex, NULL );
	
	nano_gettimeofday( &zero, NULL );
	offset = 0;
	offset_u = 0;
	g_bai = 1.0;
	nano_usleep( 100000 );
	
	pthread_create( &thread_terminal, NULL, shm_terminal, NULL );
	while( !g_shutoff )
	{
		int i;
		int rst;
		rst = 0;
		if( poll( &pfd, 1, 0 ) != 0 )
		{
			char input[256];
			if( g_shutoff )
				break;
			if( fgets( input, 256, stdin ) == 0 )
			{
				continue;
			}
			if( strcmp( input, "reset\n" ) == 0 )
			{
				struct timeval tmp;
				nano_gettimeofday( &zero, NULL );
				offset = 0;
				offset_u = 0;
				g_bai = 1.0;
				user_start.tv_sec = start;
				user_start.tv_usec = 0;
				mode = MODE_RUN;
				rst = 1;
				
				nano_gettimeofday( &tmp, NULL );
				timersub( &tmp, &zero, &tmp );
				tmp.tv_sec = ( float )tmp.tv_sec * g_bai;
				tmp.tv_usec = ( float )tmp.tv_usec * g_bai;
				timeradd( &user_start, &tmp, &user_start );
				mode = MODE_PAUSE;
			}
			else if( strcmp( input, "pause\n" ) == 0 || strcmp( input, "p\n" ) == 0 )
			{
				if( mode == MODE_RUN )
				{
					struct timeval tmp;
					
					nano_gettimeofday( &tmp, NULL );
					timersub( &tmp, &zero, &tmp );
					tmp.tv_sec = ( float )tmp.tv_sec * g_bai;
					tmp.tv_usec = ( float )tmp.tv_usec * g_bai;
					timeradd( &user_start, &tmp, &user_start );
					mode = MODE_PAUSE;
				}
				else if( mode == MODE_PAUSE )
				{
					nano_gettimeofday( &zero, NULL );
					mode = MODE_RUN;
				}
			}
			else if( strcmp( input, "help\n" ) == 0 || strcmp( input, "h\n" ) == 0 )
			{
				printf( "Commands\033[0K\n" );
				printf( "h / help  : This\033[0K\n" );
				printf( "p / pause : Pause / Resume\033[0K\n" );
				printf( "reset     : Reset\033[0K\n" );
				printf( "(number)  : Add (number) second\033[0K\n" );
				printf( "Commands while pausing\033[0K\n" );
				printf( "s(number) : (number) times high speed replay\033[0K\n" );
			}
			else if( mode == MODE_PAUSE && strstr( input, "s" ) == input )
			{
				g_bai = atof( input + 1 );
			}
			else
			{
				offset += atoi( input );
			}
		}
		for ( i = 0; i < g_num; i++ )
		{
			if( g_shmend[i] == 0 )
				break;
		}
		if( i == g_num && !rst )
		{
			// ログ終了 Reset & Pause
			struct timeval tmp;
			
			nano_gettimeofday( &tmp, NULL );
			timersub( &tmp, &zero, &tmp );
			tmp.tv_sec = ( float )tmp.tv_sec * g_bai;
			tmp.tv_usec = ( float )tmp.tv_usec * g_bai;
			timeradd( &user_start, &tmp, &user_start );
			mode = MODE_PAUSE;
		}
		for ( i = 0; i < g_num; i++ )
		{
			if( g_shmwait[i] == 1 )
				break;
		}
		if( i != g_num )
		{
			struct timeval tmp;
			if( waiting == 0 )
			{
				nano_gettimeofday( &tmp, NULL );
				timersub( &tmp, &zero, &tmp );
				tmp.tv_sec = ( float )tmp.tv_sec * g_bai;
				tmp.tv_usec = ( float )tmp.tv_usec * g_bai;
				timeradd( &user_start, &tmp, &user_start );
			}
			waiting = 1;
			nano_gettimeofday( &zero, NULL );
			nano_usleep( 10000 );
			continue;
		}
		else
		{
			waiting = 0;
		}
		if( mode == MODE_PAUSE )
		{
			nano_usleep( 10000 );
			continue;
		}
		nano_gettimeofday( &( vtime ), NULL );
		timersub( &vtime, &zero, &vtime );
		vtime.tv_usec = ( float )vtime.tv_usec * g_bai;
		vtime.tv_sec = ( float )vtime.tv_sec * g_bai;
		timeradd( &vtime, &user_start, &vtime );
		vtime.tv_usec += offset_u;
		vtime.tv_sec += offset;
		if( vtime.tv_usec >= 1000000 )
		{
			vtime.tv_sec += vtime.tv_usec / 1000000;
			vtime.tv_usec %= 1000000;
		}
		
		pthread_mutex_lock( &g_simulated_time_mutex );
		g_simulated_time = vtime;
		pthread_mutex_unlock( &g_simulated_time_mutex );
		
		nano_usleep( 1000 );
	}
	pthread_join( thread_terminal, NULL );
	tcsetattr( fileno( stdin ), TCSANOW, &old_tios );
	return 1;
}

/**
	@brief 1データ再生
	@param *data shm
 */
void *play_shm( void *arg )
{
	tfshm *shm;
	int fd;
	char filename[LIBTFSHM_NAME_LENGTH];
	size_t size = 0;
	void *data = NULL;
	struct timeval lasttime;
	struct timeval simulated_time;
	int num;
	int read_next;
	int readed;
	int pause;
	
	pthread_mutex_lock( &g_simulated_time_mutex );
	simulated_time = g_simulated_time;
	pthread_mutex_unlock( &g_simulated_time_mutex );
	shm = arg;
	snprintf( filename, LIBTFSHM_NAME_LENGTH, "%s.tfshm.log", shm->name );
	
	num = g_num;
	g_num++;
	g_shmwriten[num] = 0;
	g_shmdelay[num] = 0;
	g_shmend[num] = 0;
	g_shmwait[num] = 0;
	g_shmname[num] = shm->name;
	read_next = 1;
	readed = 0;
	pause = 0;
	timerclear( &lasttime );
	
	fd = open( filename, O_RDONLY );
	while( !g_shutoff )
	{
		tfshm_control ctl;
		struct timeval lefttime;
		if( read_next )
		{
			if( !readed )
			{
				lseek( fd, ctl.size, SEEK_CUR );
			}
			if( readn( fd, &ctl, sizeof ( ctl ) ) != 1 )
			{
				// ログ終了
				g_shmend[num] = 1;
			}
			else
			{
				g_shmwriten[num]++;
				if( ctl.size > size )
				{
					if( data )
						free( data );
					data = malloc( ctl.size );
					size = ctl.size;
				}
				readed = 0;
			}
			read_next = 0;
		}
		// 待ち
		lasttime = simulated_time;
		pthread_mutex_lock( &g_simulated_time_mutex );
		simulated_time = g_simulated_time;
		pthread_mutex_unlock( &g_simulated_time_mutex );
		if( timercmp( &simulated_time, &lasttime, < ) )
		{
			// 時間の逆走
			lseek( fd, 0, SEEK_SET );
			read_next = 1;
			readed = 1;
			g_shmend[num] = 0;
			continue;
		}
		if( g_shmend[num] )
		{
			nano_usleep( 10000 );
			continue;
		}
		if( timercmp( &simulated_time, &ctl.timestamp, < ) )
		{
			// データの出力時間になっていなかったら読み込んで待つ
			struct timeval simulated_time2;
			int usec;
			
			if( !readed )
			{
				readn( fd, data, ctl.size );
				readed = 1;
			}
			pthread_mutex_lock( &g_simulated_time_mutex );
			simulated_time2 = g_simulated_time;
			pthread_mutex_unlock( &g_simulated_time_mutex );
			if( timercmp( &simulated_time2, &ctl.timestamp, < ) )
			{
				timersub( &ctl.timestamp, &simulated_time2, &lefttime );
				if( lefttime.tv_sec > 0 )
				{
					lefttime.tv_usec = 1000000;
				}
				usec = lefttime.tv_usec - 1000;
				if( usec > 0 )
					nano_usleep( lefttime.tv_usec );
				if( lefttime.tv_sec > 0 )
					continue;
				if( timercmp( &simulated_time, &lasttime, == ) )
				{
					nano_usleep( 10000 );
					continue;
				}
			}
		}
		else
		{
			// 出力時間を過ぎていたら
			timersub( &simulated_time, &ctl.timestamp, &lefttime );
			// lefttime: 今読み込んだデータを出力するまでの時間
			if( lefttime.tv_usec > 500000 || lefttime.tv_sec > 0 )
			{
				// 500ミリ秒以上古いデータなら読み飛ばす
				read_next = 1;
				g_shmdelay[num]++;
				g_shmwait[num] = 1;
				continue;
			}
		}
		g_shmwait[num] = 0;
		if( !readed )
		{
			// 読み込みが間に合っていない
			readn( fd, data, ctl.size );
			readed = 1;
		}
		tfshm_writeex( shm, data, &ctl );
		read_next = 1;
	}
	free( data );
	close( fd );
	tfshm_close( shm );
	return NULL;
}
